@@ -3,7 +3,7 @@
 
 # "Asserting Expectations" - a chapter of "The Debugging Book"
 # Web site: https://www.debuggingbook.org/html/Assertions.html
-# Last change: 2021-03-06 16:40:22+01:00
+# Last change: 2021-05-18 11:32:43+02:00
 #
 # Copyright (c) 2021 CISPA Helmholtz Center for Information Security
 # Copyright (c) 2018-2020 Saarland University, authors, and contributors
@@ -54,7 +54,6 @@ Notably, assertions detect _violations_ of these assumptions at runtime:
 
 >>> with ExpectError():
 >>>     y = my_square_root(-1)
-
 Traceback (most recent call last):
   File "", line 2, in 
     y = my_square_root(-1)
@@ -62,12 +61,12 @@ Traceback (most recent call last):
     assert x >= 0
 AssertionError (expected)
 
+
 _System assertions_ help to detect invalid memory operations.
 
 >>> managed_mem = ManagedMemory()
 >>> managed_mem
 
-
 |Address|0|1|2|3|4|5|6|7|8|9|
 |:---|:---|:---|:---|:---|:---|:---|:---|:---|:---|:---|
 |Allocated| | | | | | | | | | |
@@ -75,10 +74,8 @@ _System assertions_ help to detect invalid memory operations.
 |Content|-1|0| | | | | | | | |
 
 
-
 >>> with ExpectError():
 >>>     x = managed_mem[2]
-
 Traceback (most recent call last):
   File "", line 2, in 
     x = managed_mem[2]
@@ -89,6 +86,7 @@ Traceback (most recent call last):
 AssertionError: Reading from unallocated memory (expected)
 
 
+
 For more details, source, and documentation, see
 "The Debugging Book - Asserting Expectations"
 at https://www.debuggingbook.org/html/Assertions.html
@@ -1315,10 +1313,28 @@ if __name__ == '__main__':
 
 
 import os
+import shutil
 
 if __name__ == '__main__':
-    os.system('rm -fr assert.h testassert* testoverflow* testuseafterfree*')
-    pass
+    for path in [
+                    'assert.h',
+                    'testassert',
+                    'testassert.c',
+                    'testassert.dSYM',
+                    'testoverflow',
+                    'testoverflow.c',
+                    'testoverflow.dSYM',
+                    'testuseafterfree',
+                    'testuseafterfree.c',
+                    'testuseafterfree.dSYM',
+                ]:
+        if os.path.isdir(path):
+            shutil.rmtree(path)
+        else:
+            try:
+                os.remove(path)
+            except FileNotFoundError:
+                pass
 
 ## Next Steps
 ## ----------
@@ -1376,6 +1392,7 @@ if __name__ == '__main__':
 
 from typing import Sequence, Any, Callable, Optional, Type, Tuple, Any
 from typing import Dict, Union, Set, List, FrozenSet, cast
+
 from types import TracebackType
 
 class Storage:
@@ -1459,4 +1476,13 @@ if __name__ == '__main__':
         print(storage['123'])
 
 if __name__ == '__main__':
-    os.remove('mydb.db');
+    try:
+        os.remove('mydb.db')  # on macOS
+    except FileNotFoundError:
+        pass
+
+if __name__ == '__main__':
+    try:
+        os.remove('mydb')  # on Linux
+    except FileNotFoundError:
+        pass
