@@ -3,7 +3,7 @@
 
 # "Where the Bugs are" - a chapter of "The Debugging Book"
 # Web site: https://www.debuggingbook.org/html/ChangeCounter.html
-# Last change: 2021-03-09 18:23:35+01:00
+# Last change: 2021-05-25 12:35:34+02:00
 #
 # Copyright (c) 2021 CISPA Helmholtz Center for Information Security
 # Copyright (c) 2018-2020 Saarland University, authors, and contributors
@@ -53,37 +53,36 @@ where `repository` is either
 * a _directory_ containing a `git` clone (i.e., it contains a `.git` directory)
 * the URL of a `git` repository.
 
-Additional arguments are being passed to the underlying `RepositoryMining` class from the [PyDriller](https://pydriller.readthedocs.io/) Python package. A `filter` keyword argument, if given, is a predicate that takes a modification (from PyDriller) and returns True if it should be included.
+Additional arguments are being passed to the underlying `Repository` class from the [PyDriller](https://pydriller.readthedocs.io/) Python package. A `filter` keyword argument, if given, is a predicate that takes a modification (from PyDriller) and returns True if it should be included.
 
 In a change counter, all elements in the repository are represented as _nodes_ – tuples $(f_1, f_2, ..., f_n)$ that denote a _hierarchy_: Each $f_i$ is a directory holding $f_{i+1}$, with $f_n$ being the actual file.
 
 A `change_counter` provides a number of attributes. `changes` is a mapping of nodes to the number of changes in that node:
 
->>> change_counter.changes[('README.md',)]
-
-11
+>>> change_counter.changes.get(('README.md',), None)
+13
 
 The `messages` attribute holds all commit messages related to that node:
 
->>> change_counter.messages[('README.md',)]
-
-['first commit',
- 'Adjusted to debuggingbook',
- 'New Twitter handle: @Debugging_Book',
+>>> change_counter.messages.get(('README.md',), None)
+['Fix: corrected rule for rendered notebooks (#24)\nNew: strip out any  tags\nNew: when rendering .md files, replace videos by proper image',
  'Doc update',
  'Doc update',
+ 'New: show badges at top of GitHub project page',
+ 'More badges',
+ 'Fix: bad links in CI badges',
+ 'New: prefer Unicode arrows over LaTeX ones',
+ 'Updated README.md',
+ 'Update',
  'Doc update',
  'Doc update',
- 'Added link to author homepage',
- 'Doc update',
  'Doc update',
  'Doc update']
 
 The `sizes` attribute holds the (last) size of the respective element:
 
->>> change_counter.sizes[('README.md',)]
-
-11221
+>>> change_counter.sizes.get(('README.md',), None)
+13025
 
 `FineChangeCounter` acts like `ChangeCounter`, but also retrieves statistics for elements _within_ the respective files; it has been tested for C, Python, and Jupyter Notebooks and should provide sufficient results for programming languages with similar syntax.
 
@@ -91,7 +90,6 @@ The `map()` method of `ChangeCounter` and `FineChangeCounter` produces an intera
 
 >>> fine_change_counter.map()
 
-
    
 
 The included classes offer several methods that can be overridden in subclasses to customize what to mine and how to visualize it. See the chapter for details.
@@ -152,13 +150,16 @@ if __name__ == '__main__':
 
 
 
-from pydriller import RepositoryMining  # https://pydriller.readthedocs.io/
+from pydriller import Repository  # https://pydriller.readthedocs.io/
+
 from pydriller.domain.commit import Commit
 
+from pydriller.domain.commit import ModifiedFile
+
 import os
 
-from typing import Sequence, Any, Callable, Optional, Type, Tuple, Any
-from typing import Dict, Union, Set, List, FrozenSet, cast
+from typing import Callable, Optional, Type, Tuple, Any
+from typing import Dict, Union, Set, List
 
 def current_repo() -> Optional[str]:
     path = os.getcwd()
@@ -178,8 +179,19 @@ def current_repo() -> Optional[str]:
 if __name__ == '__main__':
     current_repo()
 
+from datetime import datetime
+
 if __name__ == '__main__':
-    book_miner = RepositoryMining(current_repo())
+    book_miner = Repository(current_repo(), to=datetime(2020, 10, 1))
+
+DEBUGGINGBOOK_REMOTE_REPO = 'https://github.com/uds-se/debuggingbook.git'
+# book_miner = Repository(DEBUGGINGBOOK_REMOTE_REPO)
+
+if __name__ == '__main__':
+    if 'CI' in os.environ:
+        # The CI git clone is shallow, so access full repo remotely
+        book_miner = Repository(DEBUGGINGBOOK_REMOTE_REPO,
+                                to=datetime(2020, 10, 1))
 
 if __name__ == '__main__':
     book_commits = book_miner.traverse_commits()
@@ -198,28 +210,28 @@ if __name__ == '__main__':
     book_first_commit.author.name, book_first_commit.author.email
 
 if __name__ == '__main__':
-    book_first_commit.modifications
+    book_first_commit.modified_files
 
 if __name__ == '__main__':
-    [attr for attr in dir(book_first_commit.modifications[0]) if not attr.startswith('_')]
+    [attr for attr in dir(book_first_commit.modified_files[0]) if not attr.startswith('_')]
 
 if __name__ == '__main__':
-    book_first_commit.modifications[0].new_path
+    book_first_commit.modified_files[0].new_path
 
 if __name__ == '__main__':
-    print(book_first_commit.modifications[0].source_code)
+    print(book_first_commit.modified_files[0].source_code)
 
 if __name__ == '__main__':
-    print(book_first_commit.modifications[0].source_code_before)
+    print(book_first_commit.modified_files[0].source_code_before)
 
 if __name__ == '__main__':
     book_second_commit = next(book_commits)
 
 if __name__ == '__main__':
-    [m.new_path for m in book_second_commit.modifications]
+    [m.new_path for m in book_second_commit.modified_files]
 
 if __name__ == '__main__':
-    readme_modification = [m for m in book_second_commit.modifications if m.new_path == 'README.md'][0]
+    readme_modification = [m for m in book_second_commit.modified_files if m.new_path == 'README.md'][0]
 
 if __name__ == '__main__':
     print(readme_modification.source_code_before)
@@ -233,6 +245,9 @@ if __name__ == '__main__':
 if __name__ == '__main__':
     readme_modification.diff_parsed['added'][:10]
 
+if __name__ == '__main__':
+    del book_miner  # Save a bit of memory
+
 ## Counting Changes
 ## ----------------
 
@@ -242,10 +257,16 @@ if __name__ == '__main__':
 
 
 if __name__ == '__main__':
-    tuple('debuggingbook/notebooks/ChangeExplorer.ipynb'.split('/'))
+    tuple('debuggingbook/notebooks/ChangeCounter.ipynb'.split('/'))
 
 Node = Tuple
 
+from collections import defaultdict
+
+import warnings
+
+from git.exc import GitCommandError  # type: ignore
+
 class ChangeCounter:
     """Count the number of changes for a repository."""
 
@@ -259,13 +280,13 @@ class ChangeCounter:
         `filter` is a predicate that takes a modification and returns True 
         if it should be considered (default: consider all).
         `log` turns on logging if set.
-        `kwargs` are passed to the `RepositoryMining()` constructor.
+        `kwargs` are passed to the `Repository()` constructor.
         """
         self.repo = repo
         self.log = log
 
         if filter is None:
-            def filter(m: Commit) -> bool:
+            def filter(m: ModifiedFile) -> bool:
                 return True
         assert filter is not None
 
@@ -275,36 +296,39 @@ class ChangeCounter:
         # a folder f_1 holding a folder f_2 ... holding a file f_n.
 
         # Mapping node -> #of changes
-        self.changes: Dict[Node, int] = {}
+        self.changes: Dict[Node, int] = defaultdict(int)
 
-         # Mapping node -> list of commit messages
-        self.messages: Dict[Node, List[str]] = {}
+        # Mapping node -> list of commit messages
+        self.messages: Dict[Node, List[str]] = defaultdict(list)
 
         # Mapping node -> last size seen
-        self.sizes: Dict[Node, int] = {}
-
-        # All hashes already considered
-        self.hashes: Set[str] = set()
+        self.sizes: Dict[Node, Union[int, float]] = {}
 
         self.mine(**kwargs)
 
 class ChangeCounter(ChangeCounter):
     def mine(self, **kwargs: Any) -> None:
         """Gather data from repository. To be extended in subclasses."""
-        miner = RepositoryMining(self.repo, **kwargs)
+        miner = Repository(self.repo, **kwargs)
 
         for commit in miner.traverse_commits():
-            for m in commit.modifications:
-                m.hash = commit.hash
-                m.committer = commit.committer
-                m.committer_date = commit.committer_date
-                m.msg = commit.msg
+            try:
+                self.mine_commit(commit)
+            except GitCommandError as err:
+                # Warn about failing git commands, but continue
+                warnings.warn(str(err))
 
-                if self.include(m):
-                    self.update_stats(m)
+    def mine_commit(self, commit: Commit) -> None:
+        for m in commit.modified_files:
+            m.committer = commit.committer
+            m.committer_date = commit.committer_date
+            m.msg = commit.msg
+
+            if self.include(m):
+                self.update_stats(m)
 
 class ChangeCounter(ChangeCounter):
-    def include(self, m: Commit) -> bool:
+    def include(self, m: ModifiedFile) -> bool:
         """
         Return True if the modification `m` should be included
         (default: the `filter` predicate given to the constructor).
@@ -313,23 +337,27 @@ class ChangeCounter(ChangeCounter):
         return self.filter(m)
 
 class ChangeCounter(ChangeCounter):
-    def update_stats(self, m: Commit) -> None:
-        """Update counters with modification `m`. Can be extended in subclasses."""
+    def update_stats(self, m: ModifiedFile) -> None:
+        """
+        Update counters with modification `m`.
+        Can be extended in subclasses.
+        """
         if not m.new_path:
             return
 
         node = tuple(m.new_path.split('/'))
 
-        if m.hash not in self.hashes:
-            self.hashes.add(m.hash)
-            self.update_size(node, len(m.source_code) if m.source_code else 0)
-            self.update_changes(node, m.msg)
+        self.update_size(node, len(m.source_code) if m.source_code else 0)
+        self.update_changes(node, m.msg)
 
         self.update_elems(node, m)
 
 class ChangeCounter(ChangeCounter):
     def update_size(self, node: Tuple, size: int) -> None:
-        """Update counters for `node` with `size`. Can be extended in subclasses."""
+        """
+        Update counters for `node` with `size`.
+        Can be extended in subclasses.
+        """
         self.sizes[node] = size
 
 class ChangeCounter(ChangeCounter):
@@ -338,14 +366,12 @@ class ChangeCounter(ChangeCounter):
         Update stats for `node` changed with `commit_msg`.
         Can be extended in subclasses.
         """
-        self.changes.setdefault(node, 0)
         self.changes[node] += 1
 
-        self.messages.setdefault(node, [])
         self.messages[node].append(commit_msg)
 
 class ChangeCounter(ChangeCounter):
-    def update_elems(self, node: Tuple, m: Commit) -> None:
+    def update_elems(self, node: Tuple, m: ModifiedFile) -> None:
         """
         Update counters for subelements of `node` with modification `m`.
         To be defined in subclasses.
@@ -357,14 +383,34 @@ DEBUGGINGBOOK_REPO = current_repo()
 if __name__ == '__main__':
     DEBUGGINGBOOK_REPO
 
-def debuggingbook_change_counter(cls: Type) -> Any:
-    """Instantiate a ChangeCounter (sub)class `cls` with the debuggingbook repo"""
+DEBUGGINGBOOK_START_DATE: datetime = datetime(2021, 3, 1)
 
-    def filter(m: Commit) -> bool:
-        """Do not include the `docs/` directory; it only holds Web pages"""
-        return m.new_path and not m.new_path.startswith('docs/')
+NUM_WORKERS = 4  # Number of threads to be run in parallel
 
-    return cls(DEBUGGINGBOOK_REPO, filter=filter)
+def debuggingbook_change_counter(
+        cls: Type,
+        start_date: datetime = DEBUGGINGBOOK_START_DATE) -> Any:
+    """
+    Instantiate a ChangeCounter (sub)class `cls` with the debuggingbook repo.
+    Only mines changes after `start_date` (default: DEBUGGINGBOOK_START_DATE)
+    """
+
+    def filter(m: ModifiedFile) -> bool:
+        """
+        Do not include
+        * the `docs/` directory; it only holds generated Web pages
+        * the `notebooks/shared/` package; this is infrastructure
+        * the `synopsis` pictures; these are all generated
+         """
+        return (m.new_path and
+                not m.new_path.startswith('docs/') and
+                not m.new_path.startswith('notebooks/shared/') and
+                '-synopsis-' not in m.new_path)
+
+    return cls(DEBUGGINGBOOK_REPO,
+               filter=filter,
+               since=start_date,
+               num_workers=NUM_WORKERS)
 
 from .Timer import Timer
 
@@ -378,17 +424,17 @@ if __name__ == '__main__':
     list(change_counter.changes.keys())[:10]
 
 if __name__ == '__main__':
-    change_counter.changes[('Chapters.makefile',)]
+    change_counter.changes.get(('Chapters.makefile',), None)
 
 if __name__ == '__main__':
-    change_counter.messages[('Chapters.makefile',)]
+    change_counter.messages.get(('Chapters.makefile',), None)
 
 if __name__ == '__main__':
     for node in change_counter.changes:
         assert len(change_counter.messages[node]) == change_counter.changes[node]
 
 if __name__ == '__main__':
-    change_counter.sizes[('Chapters.makefile',)]
+    change_counter.sizes.get(('Chapters.makefile',), None)
 
 ## Visualizing Past Changes
 ## ------------------------
@@ -404,24 +450,38 @@ import plotly.graph_objects as go
 import math
 
 class ChangeCounter(ChangeCounter):
-    def map_node_sizes(self) -> Dict[Node, Union[int, float]]:
-        """Return a mapping of nodes to sizes. Can be overloaded in subclasses."""
-        # Default: use log scale
-        return {node: math.log(self.sizes[node]) if self.sizes[node] else 0
-                for node in self.sizes}
+    def map_node_sizes(self,scale: str = 'log') -> \
+        Dict[Node, Union[int, float]]:
+        """
+        Return a mapping of nodes to sizes.
+        Can be overloaded in subclasses.
+        """
 
-        # Alternative: use sqrt size
-        return {node: math.sqrt(self.sizes[node]) for node in self.sizes}
+        if scale == 'log':
+            # Default: use log scale
+            return {node: math.log(size+1) 
+                    for node, size in self.sizes.items()}
 
-        # Alternative: use absolute size
-        return self.sizes
+        elif scale == 'sqrt':
+            # Alternative: use sqrt size
+            return {node: math.sqrt(size)
+                    for node, size in self.sizes.items()}
+
+        elif scale == 'abs':
+            # Alternative: use absolute size
+            return self.sizes
+
+        else:
+            raise ValueError(f"Unknown scale: {scale}; "
+                             f"use one of [log, sqrt, abs]")
 
 class ChangeCounter(ChangeCounter):
     def map_node_color(self, node: Node) -> Optional[int]:
-        """Return a color of the node, as a number. Can be overloaded in subclasses."""
-        if node and node in self.changes:
-            return self.changes[node]
-        return None
+        """
+        Return a color of the node, as a number.
+        Can be overloaded in subclasses.
+        """
+        return self.changes.get(node)
 
 class ChangeCounter(ChangeCounter):
     def map_node_text(self, node: Node) -> Optional[str]:
@@ -429,9 +489,8 @@ class ChangeCounter(ChangeCounter):
         Return the text to be shown for the node (default: #changes).
         Can be overloaded in subclasses.
         """
-        if node and node in self.changes:
-            return str(self.changes[node])
-        return None
+        change = self.changes.get(node)
+        return str(change) if change is not None else None
 
 class ChangeCounter(ChangeCounter):
     def map_hoverinfo(self) -> str:
@@ -472,28 +531,27 @@ if __name__ == '__main__':
     change_counter.map()
 
 if __name__ == '__main__':
-    all_nodes = list(change_counter.changes.keys())
-    all_nodes.sort(key=lambda node: change_counter.changes[node], reverse=True)
-    [(node, change_counter.changes[node]) for node in all_nodes[:4]]
+    sorted(change_counter.changes.items(), key=lambda kv: kv[1], reverse=True)[:4]
 
 if __name__ == '__main__':
     all_notebooks = [node for node in change_counter.changes.keys()
                      if len(node) == 2 and node[1].endswith('.ipynb')]
-    all_notebooks.sort(key=lambda node: change_counter.changes[node], reverse=True)
+    all_notebooks.sort(key=lambda node: change_counter.changes[node],
+                       reverse=True)
 
 from .bookutils import quiz
 
 if __name__ == '__main__':
     quiz("Which two notebooks have seen the most changes over time?",
-        [
-            f"`{all_notebooks[0][1].split('.')[0]}`",
-            f"`{all_notebooks[3][1].split('.')[0]}`",
-            f"`{all_notebooks[1][1].split('.')[0]}`",
-            f"`{all_notebooks[2][1].split('.')[0]}`",
-        ], '[1234 % 3, 3702 / 1234]')
+         [
+             f"`{all_notebooks[i][1].split('.')[0]}`"
+             for i in [0, 3, 1, 2]
+             if i < len(all_notebooks)
+         ]
+         , '[1234 % 3, 3702 / 1234]')
 
 if __name__ == '__main__':
-    all_notebooks[0][1].split('.')[0], all_notebooks[1][1].split('.')[0]
+    [notebook[1].split('.')[0] for notebook in all_notebooks[:2]]
 
 ## Counting Past Fixes
 ## -------------------
@@ -509,15 +567,13 @@ class FixCounter(ChangeCounter):
     Fixes are all commits whose message starts with the word 'Fix: '
     """
 
-    def include(self, m: Commit) -> bool:
+    def include(self, m: ModifiedFile) -> bool:
         """Include all modifications whose commit messages start with 'Fix:'"""
         return super().include(m) and m and m.msg.startswith("Fix:")
 
 class FixCounter(FixCounter):
     def map_node_text(self, node: Node) -> str:
-        if node and node in self.messages:
-            return "<br>".join(self.messages[node])
-        return ""
+        return "<br>".join(self.messages.get(node, []))
 
     def map_hoverinfo(self) -> str:
         return 'label'
@@ -777,7 +833,7 @@ if __name__ == '__main__':
 class FineChangeCounter(ChangeCounter):
     """Count the changes for files in the repository and their elements"""
 
-    def update_elems(self, node: Node, m: Commit) -> None:
+    def update_elems(self, node: Node, m: ModifiedFile) -> None:
         old_source = m.source_code_before if m.source_code_before else ""
         new_source = m.source_code if m.source_code else ""
 
@@ -799,19 +855,19 @@ if __name__ == '__main__':
 if __name__ == '__main__':
     elem_nodes = [node for node in fine_change_counter.changes.keys()
                   if len(node) == 3 and node[1].endswith('.ipynb')]
-    elem_nodes.sort(key=lambda node: fine_change_counter.changes[node], reverse=True)
+    elem_nodes.sort(key=lambda node: fine_change_counter.changes[node],
+                    reverse=True)
     [(node, fine_change_counter.changes[node]) for node in elem_nodes[:1]]
 
 from .bookutils import quiz
 
 if __name__ == '__main__':
     quiz("Which is the _second_ most changed element?",
-        [
-            f"`{elem_nodes[3][2]}` in `{elem_nodes[3][1].split('.ipynb')[0]}`",
-            f"`{elem_nodes[1][2]}` in `{elem_nodes[1][1].split('.ipynb')[0]}`",
-            f"`{elem_nodes[2][2]}` in `{elem_nodes[2][1].split('.ipynb')[0]}`",
-            f"`{elem_nodes[0][2]}` in `{elem_nodes[0][1].split('.ipynb')[0]}`",
-        ], '1975308642 // 987654321')
+         [
+            f"`{elem_nodes[i][2]}` in `{elem_nodes[i][1].split('.ipynb')[0]}`"
+            for i in [3, 1, 2, 0]
+            if i < len(elem_nodes)
+         ], '1975308642 // 987654321')
 
 if __name__ == '__main__':
     [(node, fine_change_counter.changes[node]) for node in elem_nodes[:5]]
@@ -825,13 +881,13 @@ if __name__ == '__main__':
 
 
 if __name__ == '__main__':
-    change_counter.changes[('README.md',)]
+    change_counter.changes.get(('README.md',), None)
 
 if __name__ == '__main__':
-    change_counter.messages[('README.md',)]
+    change_counter.messages.get(('README.md',), None)
 
 if __name__ == '__main__':
-    change_counter.sizes[('README.md',)]
+    change_counter.sizes.get(('README.md',), None)
 
 if __name__ == '__main__':
     fine_change_counter.map()
@@ -879,9 +935,3 @@ if __name__ == '__main__':
 
 class FineFixCounter(FixCounter, FineChangeCounter):
     pass
-
-if __name__ == '__main__':
-    fine_fix_counter = debuggingbook_change_counter(FineFixCounter)
-
-if __name__ == '__main__':
-    fine_fix_counter.map()
