@@ -3,7 +3,7 @@
 
 # "Generalizing Failure Circumstances" - a chapter of "The Debugging Book"
 # Web site: https://www.debuggingbook.org/html/DDSetDebugger.html
-# Last change: 2021-03-09 13:02:03+01:00
+# Last change: 2021-05-18 11:35:58+02:00
 #
 # Copyright (c) 2021 CISPA Helmholtz Center for Information Security
 # Copyright (c) 2018-2020 Saarland University, authors, and contributors
@@ -58,7 +58,6 @@ Here is an example of how `DDSetDebugger` works. The concrete failing input `"ba
 >>> with DDSetDebugger(SIMPLE_HTML_GRAMMAR) as dd:
 >>>     remove_html_markup('"bar')
 >>> dd
-
 remove_html_markup(s='"')
 
 The abstract input tells us that the failure occurs for whatever opening and closing HTML tags as long as there is a double quote between them.
@@ -66,14 +65,12 @@ The abstract input tells us that the failure occurs for whatever opening and clo
 A programmatic interface is available as well. `generalize()` returns a mapping of argument names to (generalized) values:
 
 >>> dd.generalize()
-
 {'s': '"'}
 
 Using `fuzz()`, the abstract input can be instantiated to further concrete inputs, all set to produce the failure again:
 
 >>> for i in range(10):
 >>>     print(dd.fuzz())
-
 remove_html_markup(s='"1')
 remove_html_markup(s='"c*C')
 remove_html_markup(s='"')
@@ -85,6 +82,7 @@ remove_html_markup(s='"')
 remove_html_markup(s='"2')
 remove_html_markup(s='"\r~\t\r')
 
+
 `DDSetDebugger` can be customized by passing a subclass of `TreeGeneralizer`, which does the gist of the work; for details, see its constructor.
 The full class hierarchy is shown below.
 
@@ -187,8 +185,8 @@ if __name__ == '__main__':
 
 import fuzzingbook
 
-from typing import Sequence, Any, Callable, Optional, Type, Tuple, Any
-from typing import Dict, Union, Set, List, FrozenSet, cast, Generator
+from typing import Any, Callable, Optional, Type, Tuple
+from typing import Dict, Union, List, cast, Generator
 
 Grammar = Dict[str,  # A grammar maps strings...
                List[
@@ -1097,124 +1095,3 @@ if __name__ == '__main__':
 
 if __name__ == '__main__':
     failures / 1000
-
-### Exercise 2: Hierarchical Delta Debugging
-
-if __name__ == '__main__':
-    print('\n### Exercise 2: Hierarchical Delta Debugging')
-
-
-
-class TreeHDDReducer(TreeGeneralizer):
-    def _reduce(self, path: TreePath, tree: DerivationTree) -> DerivationTree:
-        """This is HDD"""
-
-        symbol, children = self.get_subtree(path, tree)
-        assert children is not None
-
-        if len(path) >= 1:
-            parent, parent_children = self.get_subtree(path[:-1], tree)
-            assert parent_children is not None
-            assert parent_children[path[-1]] == (symbol, children)
-
-            def test_children(children: List) -> None:
-                parent_children[path[-1]] = (symbol, children)  # type: ignore
-                s = tree_to_string(tree)
-                self.test(s)
-
-            with DeltaDebugger.DeltaDebugger() as dd:
-                test_children(children)
-
-            # display(display_tree(tree))
-
-            children = dd.min_args()['children']
-            parent_children[path[-1]] = (symbol, children)
-
-        assert children is not None
-        for i, child in enumerate(children):
-            self._reduce(path + [i], tree)
-
-        return tree
-
-    def reduce(self) -> DerivationTree:
-        return self._reduce([], self.tree)
-
-def bad_input_tree_hdd_reducer() -> TreeHDDReducer:
-    return TreeHDDReducer(SIMPLE_HTML_GRAMMAR, copy.deepcopy(bad_input_tree),
-                       remove_html_markup, log=True)    
-
-if __name__ == '__main__':
-    all_terminals(bad_input_tree_hdd_reducer().reduce())
-
-## Exercise 3: Reducing Trees
-## --------------------------
-
-if __name__ == '__main__':
-    print('\n## Exercise 3: Reducing Trees')
-
-
-
-class TreeReducer(TreeGeneralizer):
-    def new_min_tree(self, start_symbol: str) -> DerivationTree:
-        if self.log >= 2:
-            print(f"Creating new minimal tree for {start_symbol}")
-
-        fuzzer = GrammarFuzzer(self.grammar, start_symbol=start_symbol,
-                               min_nonterminals=0,
-                               max_nonterminals=0)
-        fuzzer.fuzz()
-        return fuzzer.derivation_tree
-
-def bad_input_tree_reducer() -> TreeReducer:
-    return TreeReducer(SIMPLE_HTML_GRAMMAR, bad_input_tree,
-                       remove_html_markup, log=2)    
-
-if __name__ == '__main__':
-    tree_to_string(bad_input_tree_reducer().new_min_tree('<start>'))
-
-class TreeReducer(TreeReducer):
-    def reduce_path(self, path: TreePath, tree: Optional[DerivationTree] = None) -> DerivationTree:
-        if tree is None:
-            tree = self.tree
-        assert tree is not None
-
-        symbol, children = tree
-
-        if not path or children is None:
-            return self.new_min_tree(symbol)
-
-        head = path[0]
-        new_children = (children[:head] +
-                        [self.reduce_path(path[1:], children[head])] +
-                        children[head + 1:])
-        return symbol, new_children
-
-if __name__ == '__main__':
-    tree_to_string(cast(TreeReducer, bad_input_tree_reducer()).reduce_path([0, 0, 1, 0]))
-
-class TreeReducer(TreeReducer):
-    def can_reduce(self, path: TreePath, tree: Optional[DerivationTree] = None) -> bool:
-        reduced_tree = self.reduce_path(path, tree)
-        if self.test_tree(reduced_tree):
-            # Failure no longer occurs; cannot reduce
-            return False
-
-        return True
-
-class TreeReducer(TreeReducer):
-    def reducible_paths(self) -> List[TreePath]:
-        return self.find_paths(self.can_reduce)
-
-if __name__ == '__main__':
-    cast(TreeReducer, bad_input_tree_reducer()).reducible_paths()
-
-class TreeReducer(TreeReducer):
-    def reduce(self) -> DerivationTree:
-        tree = self.tree
-        for path in self.reducible_paths():
-            tree = self.reduce_path(path, tree)
-
-        return tree
-
-if __name__ == '__main__':
-    all_terminals(cast(TreeReducer, bad_input_tree_reducer()).reduce())
