@@ -3,7 +3,7 @@
 
 # "How Debuggers Work" - a chapter of "The Debugging Book"
 # Web site: https://www.debuggingbook.org/html/Debugger.html
-# Last change: 2021-03-09 17:54:16+01:00
+# Last change: 2021-05-18 11:53:51+02:00
 #
 # Copyright (c) 2021 CISPA Helmholtz Center for Information Security
 # Copyright (c) 2018-2020 Saarland University, authors, and contributors
@@ -50,13 +50,11 @@ with Debugger():
 
 While running, you can enter _debugger commands_ at the `(debugger)` prompt. Here's an example session:
 
-['help', 'break 14', 'list', 'continue', 'step', 'print out', 'quit']
-
 >>> with Debugger():
 >>>     ret = remove_html_markup('abc')
-
 Calling remove_html_markup(s = 'abc')
 
+
 (debugger) help
 
 break      -- Set a breakoint in given line. If no line is given, list all breakpoints
@@ -69,10 +67,12 @@ print      -- Print an expression. If no expression is given, print all variable
 quit       -- Finish execution
 step       -- Execute up to the next line
 
+
 (debugger) break 14
 
 Breakpoints: {14}
 
+
 (debugger) list
 
    1> def remove_html_markup(s):  # type: ignore
@@ -90,20 +90,24 @@ Breakpoints: {14}
   15  
   16      return out
 
+
 (debugger) continue
 
                                          # tag = False, quote = False, out = '', c = 'a'
 14             out = out + c
 
+
 (debugger) step
 
                                          # out = 'a'
 6     for c in s:
 
+
 (debugger) print out
 
 out = 'a'
 
+
 (debugger) quit
 
 The `Debugger` class can be easily extended in subclasses. A new method `NAME_command(self, arg)` will be invoked whenever a command named `NAME` is entered, with `arg` holding given command arguments (empty string if none).
@@ -165,6 +169,7 @@ if __name__ == '__main__':
 
 
 from types import FrameType
+
 from typing import Any, Optional, Callable, Dict, List, Tuple, Set, TextIO
 
 class Debugger(Tracer):
@@ -257,9 +262,6 @@ if __name__ == '__main__':
     with Debugger():
         remove_html_markup('abc')
 
-if __name__ == '__main__':
-    assert not next_inputs()
-
 ### A Command Dispatcher
 
 if __name__ == '__main__':
@@ -389,9 +391,6 @@ if __name__ == '__main__':
     with Debugger():
         remove_html_markup('abc')
 
-if __name__ == '__main__':
-    assert not next_inputs()
-
 class Debugger(Debugger):
     def print_command(self, arg: str = "") -> None:
         """Print an expression. If no expression is given, print all variables"""
@@ -413,9 +412,6 @@ if __name__ == '__main__':
     with Debugger():
         remove_html_markup('abc')
 
-if __name__ == '__main__':
-    assert not next_inputs()
-
 if __name__ == '__main__':
     next_inputs(["print (s[0], 2 + 2)", "continue"]);
 
@@ -430,9 +426,6 @@ if __name__ == '__main__':
     with Debugger():
         remove_html_markup('abc')
 
-if __name__ == '__main__':
-    assert not next_inputs()
-
 ## Listing Source Code
 ## -------------------
 
@@ -462,9 +455,6 @@ if __name__ == '__main__':
     with Debugger():
         remove_html_markup('abc')
 
-if __name__ == '__main__':
-    assert not next_inputs()
-
 ## Setting Breakpoints
 ## -------------------
 
@@ -491,9 +481,6 @@ if __name__ == '__main__':
     with Debugger():
         remove_html_markup('abc')
 
-if __name__ == '__main__':
-    assert not next_inputs()
-
 from .bookutils import quiz
 
 if __name__ == '__main__':
@@ -537,9 +524,6 @@ if __name__ == '__main__':
     with Debugger():
         remove_html_markup('abc')
 
-if __name__ == '__main__':
-    assert not next_inputs()
-
 if __name__ == '__main__':
     quiz("What does the command `delete` (without argument) do?",
         [
@@ -596,9 +580,6 @@ if __name__ == '__main__':
     with Debugger():
         remove_html_markup('abc')
 
-if __name__ == '__main__':
-    assert not next_inputs()
-
 ### Quitting
 
 if __name__ == '__main__':
@@ -621,9 +602,6 @@ if __name__ == '__main__':
     with Debugger():
         remove_html_markup('abc')
 
-if __name__ == '__main__':
-    assert not next_inputs()
-
 ## Synopsis
 ## --------
 
@@ -636,15 +614,13 @@ if __name__ == '__main__':
     _, remove_html_markup_starting_line_number = \
         inspect.getsourcelines(remove_html_markup)
     next_inputs(["help", f"break {remove_html_markup_starting_line_number + 13}",
-                 "list", "continue", "step", "print out", "quit"]);
+                 "list", "continue", "step", "print out", "quit"])
+    pass
 
 if __name__ == '__main__':
     with Debugger():
         ret = remove_html_markup('abc')
 
-if __name__ == '__main__':
-    assert not next_inputs()
-
 from .ClassDiagram import display_class_hierarchy
 
 if __name__ == '__main__':
@@ -724,9 +700,6 @@ if __name__ == '__main__':
     with Debugger():
         remove_html_markup('abc')
 
-if __name__ == '__main__':
-    assert not next_inputs()
-
 ### Exercise 2: More Commands
 
 if __name__ == '__main__':
