@@ -3,7 +3,7 @@
 
 # "Reducing Failure-Inducing Inputs" - a chapter of "The Debugging Book"
 # Web site: https://www.debuggingbook.org/html/DeltaDebugger.html
-# Last change: 2021-03-09 17:55:04+01:00
+# Last change: 2021-05-18 12:21:03+02:00
 #
 # Copyright (c) 2021 CISPA Helmholtz Center for Information Security
 # Copyright (c) 2018-2020 Saarland University, authors, and contributors
@@ -48,10 +48,8 @@ Here is a simple example: An arithmetic expression causes an error in the Python
 
 >>> def myeval(inp: str) -> Any:
 >>>     return eval(inp)
-
 >>> with ExpectError(ZeroDivisionError):
 >>>     myeval('1 + 2 * 3 / 0')
-
 Traceback (most recent call last):
   File "", line 2, in 
     myeval('1 + 2 * 3 / 0')
@@ -60,6 +58,7 @@ Traceback (most recent call last):
   File "", line 1, in 
 ZeroDivisionError: division by zero (expected)
 
+
 Can we reduce this input to a minimum? _Delta Debugging_ is a simple and robust reduction algorithm. We provide a `DeltaDebugger` class that is used in conjunction with a (failing) function call:
 
 with DeltaDebugger() as dd:
@@ -72,7 +71,6 @@ The class automatically determines minimal arguments that cause the function to
 >>> with DeltaDebugger() as dd:
 >>>     myeval('1 + 2 * 3 / 0')
 >>> dd
-
 myeval(inp='3/0')
 
 The input is reduced to the maximum: We get the essence of the division by zero.
@@ -80,13 +78,11 @@ The input is reduced to the maximum: We get the essence of the division by zero.
 There also is an interface to access the reduced input(s) programmatically. The method `min_args()` returns a dictionary in which all function arguments are minimized:
 
 >>> dd.min_args()
-
 {'inp': '3/0'}
 
 In contrast, `max_args()` returns a dictionary in which all function arguments are maximized, but still pass:
 
 >>> dd.max_args()
-
 {'inp': '1 + 2 * 3  '}
 
 The method `min_arg_diff()` returns a triple of 
@@ -95,13 +91,11 @@ The method `min_arg_diff()` returns a triple of
 * their minimal failure-inducing difference:
 
 >>> dd.min_arg_diff()
-
 ({'inp': ' 3 '}, {'inp': ' 3 /0'}, {'inp': '/0'})
 
 And you can also access the function itself, as well as its original arguments.
 
 >>> dd.function().__name__, dd.args()
-
 ('myeval', {'inp': '1 + 2 * 3 / 0'})
 
 `DeltaDebugger` processes (i.e., minimizes or maximizes) all arguments that support a `len()` operation and that can be indexed – notably _strings_ and _lists_. If a function has multiple arguments, all arguments that can be processed will be processed.
@@ -274,7 +268,7 @@ PASS = 'PASS'
 FAIL = 'FAIL'
 UNRESOLVED = 'UNRESOLVED'
 
-from typing import Sequence, Any, Callable, Optional, Type, Tuple, Any
+from typing import Sequence, Any, Callable, Optional, Type, Tuple
 from typing import Dict, Union, Set, List, FrozenSet, cast
 
 def ddmin(test: Callable, inp: Sequence, *test_args: Any) -> Sequence:
@@ -353,8 +347,6 @@ import sys
 
 from types import FunctionType, FrameType, TracebackType
 
-import inspect
-
 from .StackInspector import StackInspector
 
 class NoCallError(ValueError):
@@ -555,7 +547,7 @@ if __name__ == '__main__':
 
 
 class CallReducer(CallCollector):
-    def __init__(self, log: Union[bool, int] = False) -> None:
+    def __init__(self, *, log: Union[bool, int] = False) -> None:
         """Initialize. If `log` is True, enable logging."""
         super().__init__()
         self.log = log
@@ -733,18 +725,21 @@ if __name__ == '__main__':
     split({1, 2, 3, 4, 5, 6, 7}, 3)
 
 class DeltaDebugger(CachingCallReducer):
-    def dd(self, var_to_be_reduced: str, fail_args: Dict[str, Any], mode: str = '-') -> \
-                Tuple[Sequence, Sequence, Sequence]:
+    def dd(self, var_to_be_reduced: str, fail_args: Dict[str, Any], 
+           *, mode: str = '-') -> Tuple[Sequence, Sequence, Sequence]:
         """General Delta Debugging.
-        fail_args - a dict of (failure-inducing) function arguments;
-        fail_args[var_to_be_reduced] - the input to apply dd on.
-        mode tells how the algorithm should operate:
-            '-': minimize input (ddmin),
-            '+': maximizing input (ddmax),
-            '+-': minimizing pass/fail difference (dd)
-        Returns a triple (pass, fail, diff) 
-        with maximized passing input, minimized failing input,
-        and their difference (elems in fail, but not in pass).
+        `var_to_be_reduced` - the name of the variable to reduce.
+        `fail_args` - a dict of (failure-inducing) function arguments, 
+            with `fail_args[var_to_be_reduced]` - the input to apply dd on.
+        `mode`- how the algorithm should operate:
+            '-' (default): minimize input (`ddmin`),
+            '+': maximizing input (`ddmax`),
+            '+-': minimizing pass/fail difference (`dd`)
+        Returns a triple (`pass`, `fail`, `diff`) with
+        * maximized passing input (`pass`), 
+        * minimized failing input (`fail`), and
+        * their difference `diff`
+          (elems that are in `fail`, but not in `pass`).
         """
         def test(c: Set) -> str:
             # Set up args
@@ -830,7 +825,16 @@ if __name__ == '__main__':
         mystery(failing_input)
 
 if __name__ == '__main__':
-    dd.dd('inp', {'inp': failing_input})
+    mystery_pass, mystery_fail, mystery_diff = dd.dd('inp', {'inp': failing_input})
+
+if __name__ == '__main__':
+    mystery_pass
+
+if __name__ == '__main__':
+    mystery_fail
+
+if __name__ == '__main__':
+    mystery_diff
 
 if __name__ == '__main__':
     with DeltaDebugger(log=2) as dd:
@@ -976,7 +980,8 @@ class DeltaDebugger(DeltaDebugger):
         if self._function is None:
             raise NoCallError("No function call observed")
         if self.exception() is None:
-            raise NotFailingError(f"{self.format_call()} did not raise an exception")
+            raise NotFailingError(
+                f"{self.format_call()} did not raise an exception")
 
         if self.log:
             print(f"Observed {self.format_call()}" +
@@ -1088,9 +1093,7 @@ def string_error(s1: str, s2: str) -> None:
 if __name__ == '__main__':
     with DeltaDebugger(log=True) as dd:
         string_error("foo", "foobar")
-    dd.min_args()
 
-if __name__ == '__main__':
     string_error_args = dd.min_args()
     string_error_args
 
@@ -1098,6 +1101,13 @@ if __name__ == '__main__':
     with ExpectError(AssertionError):
         string_error(string_error_args['s1'], string_error_args['s2'])
 
+### Invoking an Interactive Debugger
+
+if __name__ == '__main__':
+    print('\n### Invoking an Interactive Debugger')
+
+
+
 from .Debugger import Debugger  # minor dependency
 
 from .bookutils import next_inputs
@@ -1244,7 +1254,7 @@ if __name__ == '__main__':
 def compile_and_test_html_markup(lines: List[str]) -> None:
     compile_and_run(lines +
         [
-            '''''',
+            '',
             '''if remove_html_markup('<foo>bar</foo>') != 'bar':\n''',
             '''    raise RuntimeError("Missing functionality")\n''',
             '''assert remove_html_markup('"foo"') == '"foo"', "My Test"\n'''
@@ -1260,14 +1270,14 @@ if __name__ == '__main__':
     reduced_assertions_source_lines = dd.min_args()['lines']
 
 if __name__ == '__main__':
-    print_content("".join(reduced_assertions_source_lines), ".py")
+    print_content(''.join(reduced_assertions_source_lines), '.py')
 
 if __name__ == '__main__':
     len(reduced_assertions_source_lines) / len(assertions_source_lines)
 
 if __name__ == '__main__':
     remove_html_markup_source_lines, _ = inspect.getsourcelines(Assertions.remove_html_markup)
-    print_content("".join(remove_html_markup_source_lines), ".py")
+    print_content(''.join(remove_html_markup_source_lines), '.py')
 
 if __name__ == '__main__':
     quiz("In the reduced version, what has changed?",
