@@ -3,7 +3,7 @@
 
 # "Mining Function Specifications" - a chapter of "The Debugging Book"
 # Web site: https://www.debuggingbook.org/html/DynamicInvariants.html
-# Last change: 2021-03-06 16:35:41+01:00
+# Last change: 2021-05-26 17:39:21+02:00
 #
 # Copyright (c) 2021 CISPA Helmholtz Center for Information Security
 # Copyright (c) 2018-2020 Saarland University, authors, and contributors
@@ -51,7 +51,6 @@ Both work by _observing_ a function and its invocations within a `with` clause.
 
 >>> def sum2(a, b):  # type: ignore
 >>>     return a + b
-
 >>> with TypeAnnotator() as type_annotator:
 >>>     sum2(1, 2)
 >>>     sum2(-4, -5)
@@ -60,11 +59,11 @@ Both work by _observing_ a function and its invocations within a `with` clause.
 The `typed_functions()` method will return a representation of `sum2()` annotated with types observed during execution.
 
 >>> print(type_annotator.typed_functions())
-
 def sum2(a: int, b: int) ->int:
     return a + b
 
 
+
 The invariant annotator works in a similar fashion:
 
 >>> with InvariantAnnotator() as inv_annotator:
@@ -75,7 +74,6 @@ The invariant annotator works in a similar fashion:
 The `functions_with_invariants()` method will return a representation of `sum2()` annotated with inferred pre- and postconditions that all hold for the observed values.
 
 >>> print(inv_annotator.functions_with_invariants())
-
 @precondition(lambda a, b: isinstance(a, int))
 @precondition(lambda a, b: isinstance(b, int))
 @postcondition(lambda return_value, a, b: a == return_value - b)
@@ -87,6 +85,7 @@ def sum2(a, b):  # type: ignore
     return a + b
 
 
+
 Such type specifications and invariants can be helpful as _oracles_ (to detect deviations from a given set of runs). The chapter gives details on how to customize the properties checked for.
 
 
@@ -120,6 +119,9 @@ if __name__ == '__main__':
 
 from .Tracer import Tracer
 
+from typing import Sequence, Any, Callable, Tuple
+from typing import Dict, Union, Set, List, cast, Optional
+
 ## Synopsis
 ## --------
 
@@ -230,8 +232,6 @@ if __name__ == '__main__':
     with ExpectError(enforce.exceptions.RuntimeTypeError):
         square_root_with_checked_type_annotations(1)
 
-from typing import Union, Optional
-
 @enforce.runtime_validation
 def square_root_with_union_type(x: Union[int, float]) -> float:
     """Computes the square root of x, using the Newton-Raphson method"""
@@ -307,9 +307,7 @@ if __name__ == '__main__':
 
 
 
-from typing import Sequence, Any, Callable, Optional, Type, Tuple, Any
-from typing import Dict, Union, Set, List, cast, TypeVar
-from types import FrameType, TracebackType
+from types import FrameType
 
 Arguments = List[Tuple[str, Any]]
 
@@ -637,8 +635,6 @@ def annotate_function_with_types(function_name: str,
     function_ast = ast.parse(function_code)
     return annotate_function_ast_with_types(function_ast, function_calls)
 
-from typing import Any
-
 def annotate_function_ast_with_types(function_ast: ast.AST,
                                      function_calls: List[Tuple[Arguments, Any]]) -> ast.AST:
     parameter_types: Dict[str, str] = {}
