@@ -3,7 +3,7 @@
 
 # "Repairing Code Automatically" - a chapter of "The Debugging Book"
 # Web site: https://www.debuggingbook.org/html/Repairer.html
-# Last change: 2021-03-07 12:30:54+01:00
+# Last change: 2021-05-18 17:14:48+02:00
 #
 # Copyright (c) 2021 CISPA Helmholtz Center for Information Security
 # Copyright (c) 2018-2020 Saarland University, authors, and contributors
@@ -42,7 +42,7 @@ but before you do so, _read_ it and _interact_ with it at:
 
     https://www.debuggingbook.org/html/Repairer.html
 
-This chapter provides tools and techniques for automated repair of program code. The `Repairer()` class takes a `RankingDebugger` debugger as input (such as `OchiaiDebugger` from [the chapter on statistical debugging](StatisticalDebugger.ipynb). A typical setup looks like this:
+This chapter provides tools and techniques for automated repair of program code. The `Repairer` class takes a `RankingDebugger` debugger as input (such as `OchiaiDebugger` from the [chapter on statistical debugging](StatisticalDebugger.ipynb). A typical setup looks like this:
 
 from debuggingbook.StatisticalDebugger import OchiaiDebugger
 
@@ -56,7 +56,7 @@ repairer = Repairer(debugger)
 
 Here, `test_foo()` is a function that raises an exception if the tested function `foo()` fails. If `foo()` passes, `test_foo()` should not raise an exception.
 
-The `repair()` method of a `Repairer` searches for a repair of the code covered in the debugger (except for methods starting or ending in `test`, such that `foo()`, not `test_foo()` is repaired). `repair()` returns the best fix candidate as a pair `(tree, fitness)` where `tree` is a [Python abstract syntax tree](http://docs.python.org/3/library/ast) (AST) of the fix candidate, and `fitness` is the fitness of the candidate (a value between 0 and 1). A `fitness` of 1.0 means that the candidate passed all tests. A typical usage looks like this:
+The `repair()` method of a `Repairer` searches for a repair of the code covered in the debugger (except for methods whose name starts or ends in `test`, such that `foo()`, not `test_foo()` is repaired). `repair()` returns the best fix candidate as a pair `(tree, fitness)` where `tree` is a [Python abstract syntax tree](http://docs.python.org/3/library/ast) (AST) of the fix candidate, and `fitness` is the fitness of the candidate (a value between 0 and 1). A `fitness` of 1.0 means that the candidate passed all tests. A typical usage looks like this:
 
 import astor
 
@@ -78,20 +78,25 @@ def middle(x, y, z):  # type: ignore
         elif x > z:
             return x
     return z
+
 We set up a function `middle_test()` that tests it. The `middle_debugger`  collects testcases and outcomes:
 
 >>> middle_debugger = OchiaiDebugger()
-
 >>> for x, y, z in MIDDLE_PASSING_TESTCASES + MIDDLE_FAILING_TESTCASES:
 >>>     with middle_debugger:
 >>>         middle_test(x, y, z)
 
-The repairer attempts to repair the invoked function (`middle()`). The returned AST `tree` can be output via `astor.to_source()`:
+The repairer is instantiated with the debugger used (`middle_debugger`):
 
 >>> middle_repairer = Repairer(middle_debugger)
->>> tree, fitness = middle_repairer.repair()
->>> print(astor.to_source(tree), fitness)
 
+The `repair()` method of the repairer attempts to repair the function invoked by the test (`middle()`).
+
+>>> tree, fitness = middle_repairer.repair()
+
+The returned AST `tree` can be output via `astor.to_source()`:
+
+>>> print(astor.to_source(tree))
 def middle(x, y, z):
     if y < z:
         if x < z:
@@ -104,7 +109,15 @@ def middle(x, y, z):
     elif x > z:
         return x
     return z
- 1.0
+
+
+
+The `fitness` value shows how well the repaired program fits the tests. A fitness value of 1.0 shows that the repaired program satisfies all tests.
+
+>>> fitness
+1.0
+
+Hence, the above program indeed is a perfect repair in the sense that all previously failing tests now pass – our repair was successful.
 
 Here are the classes defined in this chapter. A `Repairer` repairs a program, using a `StatementMutator` and a `CrossoverOperator` to evolve a population of candidates.
 
@@ -291,8 +304,8 @@ if __name__ == '__main__':
 
 from ast import NodeVisitor
 
-from typing import Sequence, Any, Callable, Optional, Type, Tuple, Any
-from typing import Dict, Union, Set, List, FrozenSet, cast
+from typing import Any, Callable, Optional, Type, Tuple
+from typing import Dict, Union, Set, List, cast
 
 class StatementVisitor(NodeVisitor):
     """Visit all statements within function defs in an AST"""
@@ -816,13 +829,14 @@ if __name__ == '__main__':
     middle = original_middle
 
 if __name__ == '__main__':
-    quiz("Some of the lines in our fix candidate are redundant. Which are these?",
+    quiz("Some of the lines in our fix candidate are redundant. "
+         "Which are these?",
         [
-            "Line 3: `if x < y`",
-            "Line 4: `if x > z`",
-            "Line 5: `return x`",
+            "Line 3: `if x < y:`",
+            "Line 4: `if x < z:`",
+            "Line 5: `return y`",
             "Line 13: `return z`"
-        ], '[eval(chr(100 - x)) for x in [49, 50]]')
+        ], '[eval(chr(100 - x)) for x in [48, 50]]')
 
 ## Simplifying
 ## -----------
@@ -1407,23 +1421,17 @@ class DefinitionVisitor(NodeVisitor):
         self.add_definition(node)
 
 if __name__ == '__main__':
-    repairer = Repairer(middle_debugger, log=4)
+    repairer = Repairer(middle_debugger, log=1)
 
 if __name__ == '__main__':
     good_fitness = repairer.fitness(middle_tree())
     good_fitness
 
-if __name__ == '__main__':
-    assert good_fitness >= 0.99, "fitness() failed"
-
 if __name__ == '__main__':
     bad_middle_tree = ast.parse("def middle(x, y, z): return x")
     bad_fitness = repairer.fitness(bad_middle_tree)
     bad_fitness
 
-if __name__ == '__main__':
-    assert bad_fitness < 0.5, "fitness() failed"
-
 #### Repairing
 
 if __name__ == '__main__':
@@ -1909,7 +1917,8 @@ if __name__ == '__main__':
 if __name__ == '__main__':
     quiz("Why aren't single quotes handled in the solution?",
         [
-            "Because they're not important. I mean, who uses 'em anyway?",
+            "Because they're not important. "
+                "I mean, y'know, who uses 'em anyway?",
             "Because they are not part of our tests? "
                 "Let me look up how they are constructed..."
         ], 1 << 1)
@@ -1956,8 +1965,15 @@ if __name__ == '__main__':
 
 if __name__ == '__main__':
     middle_repairer = Repairer(middle_debugger)
+
+if __name__ == '__main__':
     tree, fitness = middle_repairer.repair()
-    print(astor.to_source(tree), fitness)
+
+if __name__ == '__main__':
+    print(astor.to_source(tree))
+
+if __name__ == '__main__':
+    fitness
 
 from .ClassDiagram import display_class_hierarchy
 
