@@ -3,7 +3,7 @@
 
 # "Statistical Debugging" - a chapter of "The Debugging Book"
 # Web site: https://www.debuggingbook.org/html/StatisticalDebugger.html
-# Last change: 2021-03-06 16:35:10+01:00
+# Last change: 2021-05-18 11:35:07+02:00
 #
 # Copyright (c) 2021 CISPA Helmholtz Center for Information Security
 # Copyright (c) 2018-2020 Saarland University, authors, and contributors
@@ -83,7 +83,6 @@ After collecting events, you can print out the observed events – in this case,
 
 >>> debugger.event_table(args=True, color=True)
 
-
 | `remove_html_markup` | `s='abc'` | `s='abc'` | `s='"abc"'` | 
 | --------------------- | ---- | ---- | ---- | 
 |  remove_html_markup:1 |    X |    X |    X | 
@@ -108,23 +107,22 @@ If you collected coverage with `CoverageCollector`, you can also visualize the c
 
 >>> debugger
 
-
-| `remove_html_markup` | `s='abc'` | `s='abc'` | `s='"abc"'` | 
-| --------------------- | ---- | ---- | ---- | 
-|  remove_html_markup:1 |    X |    X |    X | 
-|  remove_html_markup:2 |    X |    X |    X | 
-|  remove_html_markup:3 |    X |    X |    X | 
-|  remove_html_markup:4 |    X |    X |    X | 
-|  remove_html_markup:6 |    X |    X |    X | 
-|  remove_html_markup:7 |    X |    X |    X | 
-|  remove_html_markup:8 |    - |    X |    - | 
-|  remove_html_markup:9 |    X |    X |    X | 
-| remove_html_markup:10 |    - |    X |    - | 
-| remove_html_markup:11 |    X |    X |    X | 
-| remove_html_markup:12 |    - |    - |    X | 
-| remove_html_markup:13 |    X |    X |    X | 
-| remove_html_markup:14 |    X |    X |    X | 
-| remove_html_markup:16 |    X |    X |    X | 
+   1 def remove_html_markup(s):  # type: ignore
+   2     tag = False
+   3     quote = False
+   4     out = ""
+   5  
+   6     for c in s:
+   7         if c == '<' and not quote:
+   8             tag = True
+   9         elif c == '>' and not quote:
+  10             tag = False
+  11         elif c == '"' or c == "'" and tag:
+  12             quote = not quote
+  13         elif not tag:
+  14             out = out + c
+  15  
+  16     return out
 
 
 ### Ranking Events
@@ -132,19 +130,18 @@ If you collected coverage with `CoverageCollector`, you can also visualize the c
 The method `rank()` returns a ranked list of events, starting with the most suspicious. This is useful for automated techniques that need potential defect locations.
 
 >>> debugger.rank()
-
 [('remove_html_markup', 12),
- ('remove_html_markup', 2),
- ('remove_html_markup', 11),
- ('remove_html_markup', 6),
- ('remove_html_markup', 9),
- ('remove_html_markup', 4),
- ('remove_html_markup', 16),
- ('remove_html_markup', 13),
  ('remove_html_markup', 3),
  ('remove_html_markup', 1),
- ('remove_html_markup', 7),
+ ('remove_html_markup', 13),
+ ('remove_html_markup', 6),
+ ('remove_html_markup', 11),
+ ('remove_html_markup', 16),
+ ('remove_html_markup', 4),
+ ('remove_html_markup', 9),
+ ('remove_html_markup', 2),
  ('remove_html_markup', 14),
+ ('remove_html_markup', 7),
  ('remove_html_markup', 10),
  ('remove_html_markup', 8)]
 
@@ -208,8 +205,9 @@ if __name__ == '__main__':
 
 from .Tracer import Tracer
 
-from typing import Sequence, Any, Callable, Optional, Type, Tuple, Any
-from typing import Dict, Union, Set, List, cast, TypeVar
+from typing import Any, Callable, Optional, Type, Tuple
+from typing import Dict, Set, List, TypeVar, Union
+
 from types import FrameType, TracebackType
 
 class Collector(Tracer):
@@ -259,15 +257,21 @@ class Collector(Collector):
         self._args: Optional[Dict[str, Any]] = None
         self._argstring: Optional[str] = None
         self._exception: Optional[Type] = None
-        self.classes_to_ignore = [self.__class__]
+        self.items_to_ignore: List[Union[Type, Callable]] = [self.__class__]
 
     def traceit(self, frame: FrameType, event: str, arg: Any) -> None:
-        """Tracing function. Saves the first function and calls collect()."""
-        if 'self' in frame.f_locals:
-            for cls in self.classes_to_ignore:
-                if isinstance(frame.f_locals['self'], cls):
-                    # Do not collect our own methods
-                    return
+        """
+        Tracing function.
+        Saves the first function and calls collect().
+        """
+        for item in self.items_to_ignore:
+            if (isinstance(item, type) and 'self' in frame.f_locals and
+                isinstance(frame.f_locals['self'], item)):
+                # Ignore this class
+                return
+            if item.__name__ == frame.f_code.co_name:
+                # Ignore this function
+                return
 
         if self._function is None and event == 'call':
             # Save function
@@ -294,7 +298,10 @@ class Collector(Collector):
         return self._function
 
     def argstring(self) -> str:
-        """Return the list of arguments from the first call, as a printable string"""
+        """
+        Return the list of arguments from the first call,
+        as a printable string
+        """
         if not self._argstring:
             raise ValueError("No call collected")
         return self._argstring
@@ -329,7 +336,18 @@ class Collector(Collector):
 if __name__ == '__main__':
     with Collector() as c:
         remove_html_markup('abc')
-    c.function(), c.id()
+
+if __name__ == '__main__':
+    c.function()
+
+if __name__ == '__main__':
+    c.args()
+
+if __name__ == '__main__':
+    c.id()
+
+if __name__ == '__main__':
+    c.argstring()
 
 ### Error Prevention
 
@@ -339,12 +357,14 @@ if __name__ == '__main__':
 
 
 class Collector(Collector):
-    def add_classes_to_ignore(self, classes_to_ignore: List[Type]) -> None:
+    def add_items_to_ignore(self,
+                            items_to_ignore: List[Union[Type, Callable]]) \
+                            -> None:
         """
-        Define additional classes to ignore during collection
+        Define additional classes and functions to ignore during collection
         (typically `Debugger` classes using these collectors).
         """
-        self.classes_to_ignore += classes_to_ignore
+        self.items_to_ignore += items_to_ignore
 
 class Collector(Collector):
     def __exit__(self, exc_tp: Type, exc_value: BaseException,
@@ -368,7 +388,7 @@ if __name__ == '__main__':
 
 
 
-from types import FunctionType, FrameType
+from types import FrameType
 
 from .StackInspector import StackInspector
 
@@ -381,7 +401,9 @@ class CoverageCollector(Collector, StackInspector):
         self._coverage: Coverage = set()
 
     def collect(self, frame: FrameType, event: str, arg: Any) -> None:
-        """Save coverage for an observed event."""
+        """
+        Save coverage for an observed event.
+        """
         name = frame.f_code.co_name
         function = self.search_func(name, frame)
 
@@ -476,7 +498,7 @@ class StatisticalDebugger(StatisticalDebugger):
         """Return a collector for the given outcome. 
         Additional args are passed to the collector."""
         collector = self.collector_class(*args, **kwargs)
-        collector.add_classes_to_ignore([self.__class__])
+        collector.add_items_to_ignore([self.__class__])
         return self.add_collector(outcome, collector)
 
     def add_collector(self, outcome: str, collector: Collector) -> Collector:
@@ -489,13 +511,16 @@ class StatisticalDebugger(StatisticalDebugger):
     def all_events(self, outcome: Optional[str] = None) -> Set[Any]:
         """Return a set of all events observed."""
         all_events = set()
+
         if outcome:
-            for collector in self.collectors[outcome]:
-                all_events.update(collector.events())
+            if outcome in self.collectors:
+                for collector in self.collectors[outcome]:
+                    all_events.update(collector.events())
         else:
             for outcome in self.collectors:
                 for collector in self.collectors[outcome]:
                     all_events.update(collector.events())
+
         return all_events
 
 if __name__ == '__main__':
@@ -530,7 +555,7 @@ if __name__ == '__main__':
 
 
 if __name__ == '__main__':
-    from IPython.display import display, Markdown, HTML
+    from IPython.display import Markdown
 
 import html
 
@@ -695,7 +720,7 @@ if __name__ == '__main__':
              "One",
              "Two",
              "Three"
-         ], 'int(chr(50))')
+         ], 'len([12])')
 
 ### Collecting Passing and Failing Runs
 
@@ -751,8 +776,6 @@ def test_debugger_html_simple(debugger: T1) -> T1:
         remove_html_markup('"abc"')
     return debugger
 
-import traceback
-
 class DifferenceDebugger(DifferenceDebugger):
     def __enter__(self) -> Any:
         """Enter a `with` block. Collect coverage and outcome;
@@ -760,7 +783,7 @@ class DifferenceDebugger(DifferenceDebugger):
         and PASS if it does not.
         """
         self.collector = self.collector_class()
-        self.collector.add_classes_to_ignore([self.__class__])
+        self.collector.add_items_to_ignore([self.__class__])
         self.collector.__enter__()
         return self
 
@@ -855,60 +878,34 @@ if __name__ == '__main__':
 
 
 
-class DiscreteSpectrumDebugger(DifferenceDebugger):
-    """Visualize differences between executions using three discrete colors"""
-
+class SpectrumDebugger(DifferenceDebugger):
     def suspiciousness(self, event: Any) -> Optional[float]:
-        """Return a suspiciousness value [0, 1.0]
-        for the given event, or `None` if unknown"""
-        passing = self.all_pass_events()
-        failing = self.all_fail_events()
-
-        if event in passing and event in failing:
-            return 0.5
-        elif event in failing:
-            return 1.0
-        elif event in passing:
-            return 0.0
-        else:
-            return None
-
-    def color(self, event: Any) -> Optional[str]:
-        """Return a color for the given event."""
-        suspiciousness = self.suspiciousness(event)
-        if suspiciousness is None:
-            return None
-
-        if suspiciousness > 0.8:
-            return 'mistyrose'
-        if suspiciousness >= 0.5:
-            return 'lightyellow'
-
-        return 'honeydew'
+        """
+        Return a suspiciousness value in the range [0, 1.0]
+        for the given event, or `None` if unknown.
+        To be overloaded in subclasses.
+        """
+        return None
 
+class SpectrumDebugger(SpectrumDebugger):
     def tooltip(self, event: Any) -> str:
-        """Return a tooltip for the given event."""
-        passing = self.all_pass_events()
-        failing = self.all_fail_events()
-
-        if event in passing and event in failing:
-            return "in passing and failing runs"
-        elif event in failing:
-            return "only in failing runs"
-        elif event in passing:
-            return "only in passing runs"
-        else:
-            return "never"
+        """
+        Return a tooltip for the given event (default: percentage).
+        To be overloaded in subclasses.
+        """
+        return self.percentage(event)
 
     def percentage(self, event: Any) -> str:
-        """Return the suspiciousness for the given event as percentage string"""
+        """
+        Return the suspiciousness for the given event as percentage string.
+        """
         suspiciousness = self.suspiciousness(event)
         if suspiciousness is not None:
             return str(int(suspiciousness * 100)).rjust(3) + '%'
         else:
             return ' ' * len('100%')
 
-class DiscreteSpectrumDebugger(DiscreteSpectrumDebugger):
+class SpectrumDebugger(SpectrumDebugger):
     def code(self, functions: Optional[Set[Callable]] = None, *, 
              color: bool = False, suspiciousness: bool = False,
              line_numbers: bool = True) -> str:
@@ -972,6 +969,7 @@ class DiscreteSpectrumDebugger(DiscreteSpectrumDebugger):
 
         return out
 
+class SpectrumDebugger(SpectrumDebugger):
     def _repr_html_(self) -> str:
         """When output in Jupyter, visualize as HTML"""
         return self.code(color=True)
@@ -984,6 +982,55 @@ class DiscreteSpectrumDebugger(DiscreteSpectrumDebugger):
         """Show code as string"""
         return self.code(color=False, suspiciousness=True)
 
+class DiscreteSpectrumDebugger(SpectrumDebugger):
+    """Visualize differences between executions using three discrete colors"""
+
+    def suspiciousness(self, event: Any) -> Optional[float]:
+        """
+        Return a suspiciousness value [0, 1.0]
+        for the given event, or `None` if unknown.
+        """
+        passing = self.all_pass_events()
+        failing = self.all_fail_events()
+
+        if event in passing and event in failing:
+            return 0.5
+        elif event in failing:
+            return 1.0
+        elif event in passing:
+            return 0.0
+        else:
+            return None
+
+    def color(self, event: Any) -> Optional[str]:
+        """
+        Return a HTML color for the given event.
+        """
+        suspiciousness = self.suspiciousness(event)
+        if suspiciousness is None:
+            return None
+
+        if suspiciousness > 0.8:
+            return 'mistyrose'
+        if suspiciousness >= 0.5:
+            return 'lightyellow'
+
+        return 'honeydew'
+
+    def tooltip(self, event: Any) -> str:
+        """Return a tooltip for the given event."""
+        passing = self.all_pass_events()
+        failing = self.all_fail_events()
+
+        if event in passing and event in failing:
+            return "in passing and failing runs"
+        elif event in failing:
+            return "only in failing runs"
+        elif event in passing:
+            return "only in passing runs"
+        else:
+            return "never"
+
 if __name__ == '__main__':
     debugger = test_debugger_html(DiscreteSpectrumDebugger())
 
@@ -1045,6 +1092,9 @@ class ContinuousSpectrumDebugger(DiscreteSpectrumDebugger):
         return collectors_without_event
 
     def event_fraction(self, event: Any, category: str) -> float:
+        if category not in self.collectors:
+            return 0.0
+
         all_collectors = self.collectors[category]
         collectors_with_event = self.collectors_with_event(event, category)
         fraction = len(collectors_with_event) / len(all_collectors)
@@ -1072,7 +1122,7 @@ class ContinuousSpectrumDebugger(ContinuousSpectrumDebugger):
         if hue is None:
             return None
         return 1 - hue
-    
+
     def tooltip(self, event: Any) -> str:
         return self.percentage(event)
 
@@ -1218,6 +1268,7 @@ if __name__ == '__main__':
 
 class RankingDebugger(DiscreteSpectrumDebugger):
     """Rank events by their suspiciousness"""
+
     def rank(self) -> List[Any]:
         """Return a list of events, sorted by suspiciousness, highest first."""
 
@@ -1666,9 +1717,10 @@ if __name__ == '__main__':
                                 DifferenceDebugger.collect_fail,
                                 DifferenceDebugger.only_pass_events,
                                 DifferenceDebugger.only_fail_events,
-                                DiscreteSpectrumDebugger.code,
-                                DiscreteSpectrumDebugger.__repr__,
-                                DiscreteSpectrumDebugger._repr_html_,
+                                SpectrumDebugger.code,
+                                SpectrumDebugger.__repr__,
+                                SpectrumDebugger.__str__,
+                                SpectrumDebugger._repr_html_,
                                 ContinuousSpectrumDebugger.code,
                                 ContinuousSpectrumDebugger.__repr__,
                                 RankingDebugger.rank
@@ -1754,3 +1806,24 @@ if __name__ == '__main__':
     print('\n### Exercise 2: Statistical Dependencies')
 
 
+
+### Exercise 3: Correlating with Conditions
+
+if __name__ == '__main__':
+    print('\n### Exercise 3: Correlating with Conditions')
+
+
+
+#### Part 1: Experiment
+
+if __name__ == '__main__':
+    print('\n#### Part 1: Experiment')
+
+
+
+#### Part 2: Collecting Conditions
+
+if __name__ == '__main__':
+    print('\n#### Part 2: Collecting Conditions')
+
+
