@@ -3,7 +3,7 @@
 
 # "Tracing Executions" - a chapter of "The Debugging Book"
 # Web site: https://www.debuggingbook.org/html/Tracer.html
-# Last change: 2021-03-06 16:33:12+01:00
+# Last change: 2021-05-26 17:38:26+02:00
 #
 # Copyright (c) 2021 CISPA Helmholtz Center for Information Security
 # Copyright (c) 2018-2020 Saarland University, authors, and contributors
@@ -46,7 +46,6 @@ This chapter provides a `Tracer` class that allows to log events during program
 
 >>> with EventTracer(condition='line == 223 or len(out) >= 6'):
 >>>     remove_html_markup('foobar')
-
 ...
                                          # s = 'foobar', function = 'remove_html_markup', line = 243, tag = False, quote = False, out = 'foobar', c = 'r'
 243     for c in s:
@@ -54,11 +53,11 @@ This chapter provides a `Tracer` class that allows to log events during program
 255     return out
 remove_html_markup() returns 'foobar'
 
+
 It also allows to restrict logs to specific events. Log entries are shown only if one of the given `events` changes its value:
 
 >>> with EventTracer(events=["c == '/'"]):
 >>>     remove_html_markup('foobar')
-
 ...
 Calling remove_html_markup(s = 'foobar', function = 'remove_html_markup', line = 238)
 ...
@@ -113,7 +112,7 @@ if __name__ == '__main__':
 
 from types import FrameType, TracebackType
 
-from typing import Any, Optional, Callable, Dict, List, Type, TextIO
+from typing import Any, Optional, Callable, Dict, List, Type, TextIO, cast
 
 def traceit(frame: FrameType, event: str, arg: Any) -> Optional[Callable]:
     ...
@@ -769,8 +768,6 @@ if __name__ == '__main__':
 
 from ast import NodeTransformer, FunctionDef, fix_missing_locations, AST, Module
 
-from typing import cast
-
 if __name__ == '__main__':
     subtree_to_be_injected: AST = ast.parse("print('entering function')")
 
